/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : EdgeList.cc                                              */
/*                                                                       */
/*       Date : November  3, 1997                                        */
/* Updated at : August    3, 1998                                        */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/**
    EdgeList represents a duty which can
      1) starts from and ends at any city,
      2) no longer than the maximum duty, and
      3) not contain flying time more than the allowed limit.

    The cost of the duty is calculated as follows:
      If the total flying time is less than the paid flying time,
      then it only replects the flat cost related to any duty.
      If the total flying time is more than the paid flying time,
      then it contains flat cost as well as the cost coming from
      extra flying times as a bonus.
**/
/*************************************************************************/

#include <iostream.h>
#include <stdio.h>

#include <gVector.h> 

#include "Edge.h" 
#include "City.h" 
#include "Date.h" 
#include "EdgeList.h" 
#include "debugs.h" 
#include "AirScheduleProblem.h"
#include "commons.h" 

/*************************************************************************/
/*************************************************************************/
EdgeList::EdgeList() : gVector<EdgePtr>() 
{  myReset();  
   dCostCalculated=0;
}

/*************************************************************************/
/*************************************************************************/
EdgeList::EdgeList(int size) : gVector<EdgePtr>(size) { myReset();} 

/*************************************************************************/
/*************************************************************************/
EdgeList::EdgeList(EdgePtr edge) : gVector<EdgePtr>() 
{  
   addElement(edge); 
   myReset();  
   dCostCalculated=0;
}

/*************************************************************************/
/*************************************************************************/
EdgeList::EdgeList(EdgeList& RHS) 
{  
   copyFrom(RHS); 
   myReset(); 
   dCostCalculated=0;
} 

/*************************************************************************/
/*************************************************************************/
void EdgeList::myReset()
{
   timeCalculated = 0;
   costCalculated = 0;
   ecost = 0.0;
   totalFlyingTime = 0;
   totalOverNights = 0;
   dCost=0;
   dCostCalculated=0;

}// end of myReset()

/*************************************************************************/
/*************************************************************************/
void EdgeList::addEdge(EdgePtr edge) 
{ 
   EdgePtr fEdge = NULL;
   EdgePtr sEdge = NULL;

   //cout << " in EdgeList::addEdge() " << endl;
   if(getNumberOfEdges()>0) 
   {
      fEdge = getFirstEdge();
      sEdge = getLastEdge();

      //cout << *edge << endl; 

      if(fEdge!=NULL && edge->getEndNode()==fEdge->getStartNode()) 
	 addElementFront(edge);

      if(sEdge!=NULL && edge->getStartNode()==sEdge->getEndNode()) 
         addElement(edge);

   } else addElement(edge);

   //cout << " out of EdgeList::addEdge() " << endl;
   myReset();

}// end of addEdge() 

/*************************************************************************/
/*************************************************************************/
EdgeList::addEdgeFront(EdgePtr edge) 
{ 
   int ret=0;

   //cout << " addEdgeFront() " << endl;
   //cout << *this << endl;
   //cout << *edge << endl;
   if( edge!=NULL && ( getIndex()==0 || canFollow(edge) ) ) 
   { addElementFront(edge); ret=1; }

   myReset();

   //cout << *edge << " " << ret << endl << *this << endl;
   return ret;

}// end of addEdgeFront()

/*************************************************************************/
/*************************************************************************/
Edge *EdgeList::getFirstEdge() const { return *getFirstElement(); }

/*************************************************************************/
/*************************************************************************/
Edge *EdgeList::getLastEdge() const  { return *getLastElement(); } 

/*************************************************************************/
/*************************************************************************/
Edge *EdgeList::getNthEdge(unsigned indx) const { return *getNthElement(indx); } 
/*************************************************************************/
/*************************************************************************/
int EdgeList::getNumberOfEdges() const { return getIndex(); } 

/*************************************************************************/
/*************************************************************************/
EdgeList::canPrecede(EdgeList& RHS) { return RHS.canFollow(*this); }

/*************************************************************************/
/*************************************************************************/
EdgeList::canFollow(EdgeList& RHS) 
{
   int ret = ( getStartCityIndex() == RHS.getDestinationCityIndex()) &&
	     ( getStartDate() > RHS.getEndDate());

    return ret;

}// end of canFollow()

/*************************************************************************/
/*************************************************************************/
EdgeList::canFollow(EdgePtr RHS) 
{
   int ret = ( getIndex()==0 && RHS!=NULL );

   if( !ret && RHS!=NULL )
     ret = ( getFirstEdge()->getStartNode() == RHS->getEndNode() );

   return ret;

}// end of canFollow()

/*************************************************************************/
/*************************************************************************/
Date EdgeList::getStartDate() const
{
   Date ret;
   Edge *first = *getFirstElement();
   if(first!=NULL) ret = first->getStartDate();
   return ret;

}// end of getStartDate()

/*************************************************************************/
/*************************************************************************/
Date EdgeList::getEndDate() const
{
   Date ret;
   Edge *last = *getLastElement();
   if(last!=NULL) ret = last->getEndDate();
   return ret;

}// end of getEndDate()

/*************************************************************************/
/*************************************************************************/
EdgeList::doesStartFromBaseCity() const
{ return (getSize()>0?(*getFirstElement())->doesStartFromBaseCity():0); }

/*************************************************************************/
/*************************************************************************/
EdgeList::doesEndAtBaseCity() const
{ return (getSize()>0?(*getLastElement())->doesEndAtBaseCity():0); }

/*************************************************************************/
/** Removes the edges which ends at the given node id.  */
/*************************************************************************/
void EdgeList::removeEdgesEndAt(int indx) 
{
   Edge *edge = NULL;
   int i = getSize() - 1 ;

   //cout << " In removeEdgesEndAt() " << endl;
   while( i>=0 )
   {
      edge = getNthEdge(i);
      if( edge!=NULL &&  edge->getEndNode()==indx ) removeElementAt(i);
      i--;

   }// end of while
   //cout << " Out of removeEdgesEndAt() " << endl;
   myReset();

}// end of removeEdgesEndAt()

/*************************************************************************/
/** removes the edges which starts from the given node id.  */
/*************************************************************************/
void EdgeList::removeEdgesStartFrom(int indx) 
{
   Edge *edge = NULL;
   int i = getSize() - 1;

   //cout << " In removeEdgesStartFrom() " << endl;
   while( i>=0 )
   {
      edge = getNthEdge(i);
      if( edge!=NULL &&  edge->getStartNode()==indx ) removeElementAt(i);
      i--;

   }// end of while
   //cout << " Out of removeEdgesStartFrom() " << endl;
   myReset();

}// end of removeEdgesStartFrom()

/*************************************************************************/
/*************************************************************************/
void EdgeList::removeFromTo(int from, int to) 
{

   if( from >=0 && to< getIndex() )
   {
      myReset();
      int i=to;

      while(i>=from)
      {
	 removeElementAt(i);
	 i--;

      }// end of while

   }// end of if

}// end of removeEdgesEndAt()


/*************************************************************************/
/*
// returns -1 : if the node is not in the list
//         >=0 : index of edge
*/
/*************************************************************************/

EdgeList::doesContainNodeId(int nodeId) const
{

   int ret = -1;
   int i=0;
   Edge *edge;

   while(i<getSize() && ret==-1)
   {
      edge = getNthEdge(i);
      if( edge!=NULL && ( edge->getStartNode()==nodeId || 
	  edge->getEndNode()==nodeId ) ) ret = i;
      else i++;

   }// end of while

   return ret;

}// end of doessContainNodeId()

/*************************************************************************/
/*
// returns -1 : if the node is not in the list
//         >=0 : index of edge
*/
/*************************************************************************/

void EdgeList::splitListFromNode(int nodeId, EdgeList& first, EdgeList&second ) 
{
   EdgeList _first;
   EdgeList _second;
   int i=0,j,stay=0;
   Edge *edge;

   if((i=doesContainNodeId(nodeId))>=0 && i<getIndex() )
   {
     j=0;stay=1;
     while(j<i&&stay) 
     {  
	edge = getNthEdge(j);

        if( edge!=NULL ) 
        { 
	  if( edge->getStartNode()!=nodeId && edge->getEndNode()!=nodeId ) 
	      first.addElement(edge);
          else stay = 0;

        }// end of if

	j++;

     }// end of while

     j++;

     while(j<getIndex()) 
     {  
	edge = getNthEdge(j);

        if( edge!=NULL ) second.addElement(edge);

	j++;

     }// end of while

   }// end of if

   first = _first;
   second = _second;

}// end of splitListFromNode()

/*************************************************************************/
/*
//
// Returns the number of Overnight stays. Overnight stay is decided if
// there are more than 8 hours between assignment.
// Note: This can be send as a parameter to this method.
//
*/
/*************************************************************************/

EdgeList::getNumberOfOvernightStays() 
{

   if(!timeCalculated) CalculateTime();

   return totalOverNights;

}// end of getNumberOfOvernightStays()

/*************************************************************************/
/**  Returns the total flying time of the edge list(duty period) */
/*************************************************************************/
EdgeList::getTotalFlyingTime() 
{
   if(DEBUG5) cout << " in EdgeList::getTotalFlyingTime() " << endl;

   if(!timeCalculated) CalculateTime();

   return totalFlyingTime;

}// end of getTotalFlyingTime()

/*************************************************************************/
/**  Returns the total flying time of the edge list(duty period) */
/*************************************************************************/
void EdgeList::CalculateTime() 
{
   int retVal = 0;
   int i=0;

   Flight *f;
   Edge   *edge;
   Edge   *first;
   Edge   *second;
   int    diff;
   int    ostay=0;

   if(DEBUG5) 
     cout << " in EdgeList::getTotalFlyingTime() " << endl;
   for(i=0;i<getNumberOfEdges();i++)
   {
      second = edge = getNthEdge(i);

      if(edge!=NULL)
      {
         f = edge->getFirstFlight();

         if(f!=NULL) 
	 { 
	   diff    = (f->getArrivalDate()-f->getDepartureDate());
	   retVal += (diff>0?diff:-1*diff);

	 }// end of if

      }// end of if

      if(i)
      {
	diff = second->getEndDate()-first->getStartDate();
	if( diff>8*60 ) ostay++;
      }// end of if

      first = second;

   }// end of for

   if(edge!=NULL) f = edge->getSecondFlight();

   if(f!=NULL) retVal += (f->getArrivalDate()-f->getDepartureDate());

   totalFlyingTime = retVal;
   totalOverNights = ostay;
   timeCalculated  = 1;

}// end of CalculateTime()

/*************************************************************************/
/** Calculates the Cost of this edge list.  */
/*************************************************************************/
void EdgeList::CalculateCost(float CostPerMinute,float CostPerOvernight) 
{
   float retVal = 0;
   int i;
   Edge *curr; 

   for(i=0;i<getSize();i++) 
   { curr = getNthEdge(i); retVal += (CostPerMinute*curr->getCost()); }

   retVal += (CostPerOvernight*getNumberOfOvernightStays() );

   ecost = retVal;
   costCalculated=1;

}// end of getCost()

/*************************************************************************/
/** Calculates the Cost of this edge list.  */
/*************************************************************************/
float EdgeList::getCost(float CostPerMinute,float CostPerOvernight) 
{

   if(!costCalculated) CalculateCost(CostPerMinute,CostPerOvernight);

   return ecost;

}// end of getCost()

/*************************************************************************/
/**  Returns the index of the starting city of this edge list.  */
/*************************************************************************/
EdgeList::getStartCityIndex() const
{
   int ret = -1;
   Edge *first = *getFirstElement();

   if(first!=NULL) ret = first->getStartCityIndex();

   return ret;

}// end of getStartCityIndex()

/*************************************************************************/
/**  Returns the index of destination city of this edge list.  */
/*************************************************************************/
EdgeList::getDestinationCityIndex() const
{
   int ret = -1;
   Edge *last = *getLastElement();

   if(last!=NULL) ret = last->getDestinationCityIndex();

   return ret;

}// end of getDestinationCityIndex() 

/*************************************************************************/
/**  Equality can be done comparing each element since EdgeList keeps    */
/**  them in time order.                                                 */
/*************************************************************************/
EdgeList::operator==(const EdgeList& RHS)
{
   int i, ret = 0;

   ret = (getSize() == RHS.getSize());

   for(i=0;ret && i<getSize();i++) 
       ret = ( *getNthEdge(i) == *RHS.getNthEdge(i));

   return ret;

}// end of operator==()

/*************************************************************************/
/**  Assignment operation.  */
/*************************************************************************/
EdgeList EdgeList::operator=(const EdgeList& RHS)
{
   dispose();
   copyFrom(RHS);

   myReset();
   dCostCalculated=0;

   return (*this);

}// end of operator=()

/*************************************************************************/
/*************************************************************************/
void EdgeList::nicePrint(ostream& os) const
{
   int i;
   Edge *curr = NULL; 

   //cout << " in operator<<() of Edge " << getIndex() << endl;

   os << " { ";

   for(i=0;i<getNumberOfEdges();i++) 
   { 
      curr = getNthEdge(i); 
      if(curr!=NULL) os << curr->getStartNode() << " -> " ; 
      else os << " Node is NULL " << endl;

   }// end of for

   if(curr!=NULL) os << curr->getEndNode();

   os << " } " ; //<< endl;

   os << " [ ("<<getNumberOfEdges()<<") ";

   for(i=0;i<getNumberOfEdges();i++) 
   { 
      curr = getNthEdge(i); 
      if(curr!=NULL) { curr->nicePrint2(os); os << ";" ; } 
      else os << " NULL " ;

   }// end of for

   //if(curr!=NULL) { curr->nicePrint2(os); os << ";" ; } 
   //else os << " NULL " ;

   os << " ] " << endl;

   //cout << " out of nicePrint() of EdgeList "  << endl;

}// end of  nivePrint()

/*************************************************************************/
/*************************************************************************/
ostream& operator<<(ostream& os, const EdgeList& RHS)
{
   RHS.nicePrint(os);
   return os;

}// end of  operator<<()

/*************************************************************************/
/**
    1. Total length should be less than maximum rotation time
    2. It should end where it started
    3. It should start from base city
    4. Minimum time between flights
    5. calculates the length of the first duty and the last duty.
    6. calculates the number of duties.
*/
/*************************************************************************/
EdgeList::isValid(AirScheduleProblem *problem) 
{

   int onights = 0;
   int extra   = 0;
   int perdiem = 0;
   dCost = 0; // dolar($) cost
   lastDutyLength = 0;
   float minRqRest;

   int i;
   int ret= (getEndDate()-getStartDate()<=problem->getMaxRotationPeriod());
   Edge *curr = NULL; 

   Flight *prev;
   Flight *cur;
   unsigned dutyperiod;
   unsigned tflyingtime=0;
   Date     dutyStart;
   int      flength;

   float    mxDutyTime = problem->getMaxDutyPeriod();
   float    mxFlyTime  = problem->getMaxFlyTime();
   float    miFlyTime  = problem->getMinPaidFlyTime();
   float    mincost    = problem->getMinSalary();
   float    minrest    = problem->getMinRestTime();

   unsigned NumFlights = getNumberOfEdges();
   i=0;
   curr = getNthEdge(i); 

   prev = curr->getFirstFlight();

   // Update the start city index
   int startCityIndex = prev->getFromIndex();

   // Update the flying time for the first flight
   tflyingtime = prev->getArrivalDate()-prev->getDepartureDate();

   //Update first duty length
   firstDutyLength = prev->getArrivalDate()-dutyStart;
   numberOfDuties = 1;

   // Update the start date for the rotation
   dutyStart = prev->getDepartureDate();

   while( ret && i < NumFlights )
   {  
      curr = getNthEdge(i); 
      cur = curr->getSecondFlight();
         //cur = myProblem->getFlightWithIndex(getFlightIndex(i+1)-1);
	 flength = cur->getArrivalDate()-cur->getDepartureDate();

         if( ( cur->getArrivalDate()-dutyStart)<=mxDutyTime  && 
	     ( tflyingtime+flength)<=mxFlyTime)
	 {
	     tflyingtime+=flength; // increase flying time
	     if(numberOfDuties==1) 
                firstDutyLength = cur->getArrivalDate()-dutyStart;
	 }
	 else // we should define another duty period for this guy
	   {
	     /*
             if( ( cur->getArrivalDate()-dutyStart)>mxDutyTime  )
		 cout << " long duty " << (cur->getArrivalDate()-dutyStart) 
		      << " max " << mxDutyTime
		      << endl; 
	     if( ( tflyingtime+flength)>mxFlyTime )
		 cout << " long flying " << (tflyingtime+flength) << endl; 
             */		  

	      // Find minimum rest between duties 
	      minRqRest = problem->getMinimumRestForDuty(prev->getArrivalDate()-dutyStart);
	      // Minimum rest between duties should be larger than
	      // required minimum rest time.
	      if( (cur->getDepartureDate()-prev->getArrivalDate()) >= minRqRest) 
	      {

		 // update cost at least minimum payment for this duty
		 dCost += mincost;

		 // update over flying time
		 if(tflyingtime-miFlyTime>0) extra+=(tflyingtime-miFlyTime);

		 // update total flying time for the next duty
		 tflyingtime = flength;

		 numberOfDuties++;

		 // update start time for the next duty
		 dutyStart = cur->getDepartureDate();

		 // if crew stays in another city
		 if( prev->getToIndex() != startCityIndex )
		 {
		    // update overnight stay
		    onights++;

		    // update per diem expense
		    perdiem += (cur->getDepartureDate()-prev->getArrivalDate());

		 }// end of if

	      } else { 
		       ret=0; // it is not feasible
		       /*cout << " not enough rest between " << endl
			    << *prev << endl
			    << *cur << endl;

                       cout << " tduty " 
			    << ( prev->getArrivalDate()-dutyStart)
			    << " > mxDutyTime " << mxDutyTime
		            << " minRqRest " << minRqRest
		            << " duty has " 
			    << (cur->getDepartureDate()-prev->getArrivalDate())
		            << " mxFlyTime " << mxFlyTime
		            << " cFlyTime " << tflyingtime
		            << " wbFlyTime " << (tflyingtime+flength)
		            << endl;*/
		     } 
	   }

         prev = cur; 

         i++; 

   }// end of while

   // last duty
   if(ret)
   {
      // update cost at least minimum payment for this duty
      dCost += mincost;

      // update over flying time
      if(tflyingtime-miFlyTime>0) extra+=(tflyingtime-miFlyTime);

      // update the length of last duty
      lastDutyLength = cur->getArrivalDate()-dutyStart;

   }// end of if

   // if this is feasible, then we need to add the extras to the total
   // cost function.
   if(ret)
   {
      // Add Per diem expenses
      if(perdiem) dCost += (perdiem*problem->getPerDiemPerMinute()); 

      // Add hotel expenses
      if(onights) dCost += (onights*problem->getHotelCost());

      // Add bonus for extra flying times
      if(extra) dCost += (extra*problem->getBonusForExtraFlyMinute());

      //if(getStartCityIndex()!=getDestinationCityIndex()) dCost *= problem->();
   }// end of if

   dCostCalculated=1;

   //if(!ret) cout << " notValid " ;

   return ret;

}// end of isValid()

/*************************************************************************/
/** Calculates the Cost of this edge list.  */
/*************************************************************************/
float EdgeList::getDCost( AirScheduleProblem *problem )
{

   //cout << " ELgetDCost() ";
   if(!dCostCalculated) isValid(problem);

   return dCost;

}// end of getDCost()

/*************************************************************************/
/** Returns the length of the last duty. */
/*************************************************************************/
float EdgeList::getLastDutyLength( AirScheduleProblem *problem )
{

   if(!dCostCalculated) isValid(problem);

   return lastDutyLength;

}// end of getLastDutyLength()

/*************************************************************************/
/** Returns the length of the first duty. */
/*************************************************************************/
float EdgeList::getFirstDutyLength( AirScheduleProblem *problem )
{

   if(!dCostCalculated) isValid(problem);

   return firstDutyLength;

}// end of getFirstDutyLength()

/*************************************************************************/
/** */
/*************************************************************************/
EdgeList::canMerge( EdgeList& eList, AirScheduleProblem* problem)
{

   int i;
   Date startD;  
   Date endD;  
   int ret=0;
   if( eList.getEndDate()<getStartDate() ) 
   {
       //cout << *this << endl << eList << endl;
       //cout << " it can be added infront " << endl;
       startD=eList.getStartDate();
       endD=getEndDate();

       if((endD-startD)<=problem->getMaxRotationPeriod())
       {

          Edge *last = eList.getLastEdge();
          Edge *first = getFirstEdge();

          // If there is an edge to combine them
          if( last!=NULL && first!=NULL && 
	      last->getEdgeGraph()->getEdgeFromTo(last->getEndNode(),first->getStartNode())!=NULL) 
	  { 
              EdgeList a(eList.getSize()+getSize());
              for(i=0;i<eList.getSize();i++) a.addEdge(eList.getNthEdge(i));
	      a.addEdge(last->getEdgeGraph()->getEdgeFromTo(last->getEndNode(),first->getStartNode()));
              for(i=0;i<getSize();i++) a.addEdge(getNthEdge(i));
              ret = a.isValid(problem);

	      //if(ret) cout << a << " is valid " << endl;
	      //else  cout << a << " is *NOT* valid " << endl;

          }// end of if

       }// end of if

   }
   else
   if( getEndDate()<eList.getStartDate())
   {
       //cout << " it can be added to end " << endl;
       startD=getStartDate();
       endD=eList.getEndDate();
       if( (endD-startD)<=problem->getMaxRotationPeriod())
       {
          Edge *last = getLastEdge();
          Edge *first = eList.getFirstEdge();

	  // If there is an edge to combine them 
          if( last!=NULL && first!=NULL && 
	      last->getEdgeGraph()->getEdgeFromTo(last->getEndNode(),first->getStartNode())!=NULL) 
          {
             EdgeList a(eList.getSize()+getSize());
             for(i=0;i<getSize();i++) a.addEdge(getNthEdge(i));
             a.addEdge(last->getEdgeGraph()->getEdgeFromTo(last->getEndNode(),first->getStartNode()));
             for(i=0;i<eList.getSize();i++) a.addEdge(eList.getNthEdge(i));
             ret= a.isValid(problem);

	     //if(ret) cout << a << " is valid " << endl;
	     //else  cout << a << " is *NOT* valid " << endl;

          }// end of if

       }// end of if
   }

   return ret;
      
}// end of canMerge()

/*************************************************************************/
/** */
/*************************************************************************/
EdgeList::mergeWith( EdgeList& eList )
{

   int i;
   Date startD;  
   Date endD;  
   int ret=0;

   if( eList.getEndDate()<getStartDate() ) 
   {
       EdgeList a(eList.getSize()+getSize());
       for(i=0;i<eList.getSize();i++) a.addEdge(eList.getNthEdge(i));
       Edge *last = eList.getLastEdge();
       Edge *first = getFirstEdge();
       a.addEdge(last->getEdgeGraph()->getEdgeFromTo(last->getEndNode(),first->getStartNode()));
       //cout << " a " << a << endl;
       for(i=0;i<getSize();i++) a.addEdge(getNthEdge(i));
       //cout << " this " << *this << endl;
       //cout << " eList " << eList << endl;
       //cout << " a " << a << endl;
       *this=a;
       //cout << " after merger " << *this << endl;
       ret= 1;


   }
   else
   if( getEndDate()<eList.getStartDate())
   {
       EdgeList a(eList.getSize()+getSize());
       for(i=0;i<getSize();i++) a.addEdge(getNthEdge(i));
       //cout << " a " << a << endl;
       Edge *last = getLastEdge();
       Edge *first = eList.getFirstEdge();
       a.addEdge(last->getEdgeGraph()->getEdgeFromTo(last->getEndNode(),first->getStartNode()));
       for(i=0;i<eList.getSize();i++) a.addEdge(eList.getNthEdge(i));
       //cout << " this " << *this << endl;
       //cout << " eList " << eList << endl;
       //cout << " a " << a << endl;
       *this=a;
       //cout << " after merger " << *this << endl;
       ret= 1;
   }

   if(ret)
   {
      dCostCalculated=0;
      //cout << " after merger " << *this << endl;
      //char aaa; cin >> aaa;
   }

   return ret;

}// end of mergeWith()

/*************************************************************************/
/** 
    Returns the Perminute cost of uncovered flights.
**/
/*************************************************************************/
float 
EdgeList::getCostPerUncoveredMinute(AirScheduleProblem* myProblem, char* covered)
{
   float ret = MAX_FLOAT;

   int i=0;
   Edge *edge;
   float minutes = 0;

   while(i<getSize())
   {
      edge = getNthEdge(i);
      if( edge!=NULL && covered[edge->getStartNode()]==0 )
	  minutes += myProblem->getFlightWithIndex(edge->getStartNode())->getLength();

      i++;

   }// end of while

   if( edge!=NULL && covered[edge->getEndNode()]==0 )
       minutes += myProblem->getFlightWithIndex(edge->getEndNode())->getLength();

   if(minutes>0) ret = getDCost(myProblem)/minutes;

   return ret;

}// end of getCostPerUncoveredMinute()

/*************************************************************************/
/** 
    Returns the number of different elements.
**/
/*************************************************************************/
EdgeList::getNumberOfDifferences(EdgeListPtr eList)
{
   int i, j, ret = 0;
   int *nodes  = new int[getSize()+1];
   int *nodes1 = new int[eList->getSize()+1];
   Edge *edge=NULL;

   for(i=0;i<getSize();i++)
   {
      edge = getNthEdge(i);
      if( edge!=NULL ) nodes[i]=edge->getStartNode();
   }// end of for
   if( edge!=NULL ) nodes[i]=edge->getEndNode();

   edge=NULL;
   for(i=0;i<eList->getSize();i++)
   {
      edge = eList->getNthEdge(i);
      if( edge!=NULL ) nodes1[i]=edge->getStartNode();
   }// end of for
   if( edge!=NULL ) nodes1[i]=edge->getEndNode();

   i=j=0;

   while( i<=getSize() && j<=eList->getSize() )
   {
      if( nodes[i]<nodes1[j] ) { ret++; i++; }
      else if( nodes[i]>nodes1[j] ) j++; 
      else { i++; j++; }

   }// end of while

   ret += (getSize()-i+1);

   delete[] nodes;
   delete[] nodes1;

   return ret;

}// end of getNumberOfDifferences()

